/*
 * Copyright [yyyy] [name of copyright owner]
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


// Anonymous function start
//
(function( window, undefined )
{

// reference
var Config			= window.h5glib.Config;
var Debug			= window.h5glib.Debug;
var Command			= window.h5glib.Command;
var Task			= window.h5glib.Task;
var SoundTask		= window.h5glib.SoundTask;
var ReadyTask		= window.h5glib.ReadyTask;
var ImageAnimator	= window.h5glib.ImageAnimator;
var Message			= window.h5glib.Message;
var MessageHandler	= window.h5glib.MessageHandler;
var MessageManager	= window.h5glib.MessageManager;
var SceneStatus		= window.h5glib.SceneStatus;
var Scene			= window.h5glib.Scene;

var PI		= Math.PI;
var PI2		= Math.PI * 2;
var DEG_PI2	= 360;

/**
 * Priority
 */
/*
var Priority =
{
	NONE	: 0,
	STAGE	: 1,
	ACTOR	: 2,
	EVENT	: 3,
	FRONT	: 4,
	DEBUG	: 5
};
*/
/**
 * RenderInfo
 */
var RenderInfo =
{
	fov			: ( 60 * PI / 180 ),	// radian
	rayPixels	: 2,	//  8,   4,   2,   1
	samples		: 200,	// 50, 100, 200, 400
	chipPixels	: 64
};

/**
 * ProtocolInfo
 */
var ProtocolInfo =
{
	url				: "/JsFpsServer",
	method			:
	{
		join		: "join",
		update		: "update"
	},
	updateInterval	: 2000,
	playTimeMax		: 5 * 60 * 1000
};

/**
 * Res
 */
var Res =
{
	String :
	{
		HTML_USAGE		: "<p>" +
						  "＜キー入力ができない場合はゲームの画面をクリックしてみてください＞<br>" +
						  "[←][→]: 回転, [↑][↓]: 移動, [Z]: 発砲, [E]: マップエディタ起動 <br>" +
						  "キャラに接近するとイベント発動 <br>" +
						  "（馬に接近すると戦闘、猫に接近するとステージ遷移）" +
						  "</p>",
		HTML_USAGE_N	: "<p>" +
						  "＜キー入力ができない場合はゲームの画面をクリックしてみてください＞<br>" +
						  "[←][→]: 回転, [↑][↓]: 移動 <br>" +
						  "[Send] ボタン: 発言 <br>" +
						  "</p>",
		EDITOR_NAME		: "editor",
		BATTLE_NAME		: "battle",

		INPUT_HTML		: '<input id="text" type="text" size="60" maxlength="60">' +
						  '<input id="send" type="button" value="Send">',
		ID_MESSAGE		: "message",
		ID_TEXT			: "text",
		ID_SEND			: "send",

		TM_CONN_ERR		: "ゲームサーバとの接続に失敗しました。",
		TM_JOIN_NG		: "ゲームサーバへのログインが失敗しました。",
		TM_UPDATE_NG	: "ゲームサーバへの更新が失敗しました。"
	},
	Color :
	{
		TEXTBOX			: "rgba(192, 80, 77, 0.7)",
		BORDER_TEXTBOX	: "rgb(192, 192, 192)",
		FONT_TEXTBOX	: "rgb(255, 255, 255)"
	},
	Font :
	{
		EXLARGE			: "bold 24px 'ＭＳ Ｐゴシック'",
		LARGE			: "bold 18px 'ＭＳ Ｐゴシック'",
		SMALL			: "bold 14px 'ＭＳ Ｐゴシック'"
	}
};

/**
 * MessageType
 */
var MessageType =
{
	NONE			: 0,
	RET_PREV_SCENE	: 1,
	START_BATTLE	: 2,
	WARP_SCENE		: 3,
	START_EDITOR	: 4
};

/**
 * EventType
 */
var EventType =
{
	NONE	: "none",
	TEXT	: "text",
	BATTLE	: "battle",
	WARP	: "warp"
};

/**
 * StageTask
 */
var StageTask = function( scene )
{
	this.command		= scene.command;	// input

	this.map			= null;
	this.getMapHeight	= function() { return this.map.length; };
	this.getMapWidth	= function() { return this.map[0].length; };

	var RaycastingData	= function()
	{
		this.dist	= 0;
		this.val	= 0;
		this.tx		= 0;
	};
	this.rd	= [];
	for ( var i = 0; i < RenderInfo.samples; i++ )
	{
		this.rd[i]	= new RaycastingData();
	}
};
StageTask.prototype = new Task();

(function( pt )
{
	/**
	 * 
	 */
	pt.setData = function( scene, map )
	{
		this.map	= map;
	};
	/**
	 * 
	 */
	pt.draw = function( scene )
	{
		this.castRay( scene );

		var context		= scene.context;
		var canvas		= context.canvas;
		var resource	= scene.data.resource;
		var viewPoint	= scene.viewPoint;

		var image;
		// sky
		image		= resource.image.sky.data;
		var skyW2	= image.width - canvas.width;
		context.drawImage(
			image,
			Math.floor( skyW2 - ( viewPoint.ad.dir / DEG_PI2 * skyW2 ) ),	// sx
			0,				// sy
			canvas.width,	// sw
			image.height,	// sh
			0,				// dx
			0,				// dy
			canvas.width,	// dw
			image.height	// dh
		);
		// floor
		image = resource.image.floor.data;
		context.drawImage(
			image,
			0,				// sx
			0,				// sy
			canvas.width,	// sw
			image.height,	// sh
			0,				// dx
			canvas.height - image.height,		// dy
			canvas.width,	// dw
			image.height	// dh
		);

		// wall
		var radian		= viewPoint.getDirRad();
		var startTheta	= radian + ( RenderInfo.fov / 2 );

		for ( var i = 0; i < RenderInfo.samples; i++ )
		{
			var theta = startTheta - ( RenderInfo.fov * i / RenderInfo.samples );
			theta %= PI2;

			var sx	= this.rd[i].tx * ( RenderInfo.chipPixels - 1 );
			var sy	= ( this.rd[i].val - 1 ) * RenderInfo.chipPixels;
			var sw	= 1;
			var sh	= RenderInfo.chipPixels;

			var dist= this.rd[i].dist * Math.cos( theta - radian );
			var dh	= canvas.height / dist;
			var dx	= i * RenderInfo.rayPixels;
			var dy	= ( canvas.height / 2 ) - ( dh / 2 );
			var dw	= RenderInfo.rayPixels;

			context.drawImage( resource.image.wall.data, sx, sy, sw, sh, dx, dy, dw, dh );
		}
	};
	/**
	 * 
	 */
	pt.castRay = function( scene )
	{
		var viewPoint	= scene.viewPoint;
		var radian		= viewPoint.getDirRad();
		var startTheta	= radian + ( RenderInfo.fov / 2 );

		for ( var i = 0; i < RenderInfo.samples; i++ )
		{
			var theta = startTheta - ( RenderInfo.fov * i / RenderInfo.samples );
			theta %= PI2;

			var mx = Math.floor( viewPoint.ad.x );
			var my = Math.floor( viewPoint.ad.y );

			var cosTheta = Math.cos( theta );
			var sinTheta = Math.sin( theta );
			if ( cosTheta == 0 ) { cosTheta = 0.001; }
			if ( cosTheta == 0 ) { cosTheta = 0.001; }

			var distX, distY;
			var stepX, stepY;

			if ( cosTheta > 0 )
			{
				stepX = 1;
				distX = ( mx + 1 - viewPoint.ad.x ) / cosTheta;
			}
			else
			{
				stepX = -1;
				cosTheta *= -1.0;
				distX = ( viewPoint.ad.x - mx ) / cosTheta;
			}

			if ( sinTheta > 0 )
			{
				stepY = -1;
				distY = ( viewPoint.ad.y - my ) / sinTheta;
			}
			else
			{
				stepY = 1;
				sinTheta *= -1.0;
				distY = ( my + 1 - viewPoint.ad.y ) / sinTheta;
			}

			while ( true )
			{
				if ( distX < distY )
				{
					mx += stepX;
					var mapVal = ( 0 <= mx || mx < this.getMapWidth() ) ? this.map[ my ][ mx ] : 1;
					if ( mx < 0 || this.getMapWidth() <= mx || mapVal )
					{
						// not actor
						if ( mapVal < Config.actorValBase )
						{
							this.rd[i].dist	= distX;
							this.rd[i].val	= mapVal;
							this.rd[i].tx	= 1 - ( viewPoint.ad.y + distX * sinTheta * stepY ) % 1;
							break;
						}
					}
					distX += ( 1 / cosTheta );	// + dist per grid
				}
				else
				{
					my += stepY;
					var mapVal = this.map[ my ][ mx ];
					var mapVal = ( 0 <= my || my < this.getMapHeight() ) ? this.map[ my ][ mx ] : 1;
					if ( my < 0 || this.getMapHeight() <= my || mapVal )
					{
						// not actor
						if ( mapVal < Config.actorValBase )
						{
							this.rd[i].dist	= distY;
							this.rd[i].val	= mapVal;
							this.rd[i].tx	= ( viewPoint.ad.x + distY * cosTheta * stepX ) % 1;
							break;
						}
					}
					distY += ( 1 / sinTheta );	// + dist per grid
				}
			}
		}
	};
})( StageTask.prototype );

/**
 * ActorListTask
 */
var ActorListTask = function( scene )
{
	this.command	= scene.command;	// input
	this.visible	= [];
	this.maxId		= 0;
};
ActorListTask.prototype = new Task();

(function( pt )
{
	/**
	 * 
	 */
	pt.setData = function( scene, actors )
	{
		this.child	= null;

		// load actorList
		for ( var i = 0; i < actors.length; i++ )
		{
			var task	= new ActorTask();
			task.init( scene, i, actors[i] );
			// add
			if ( this.child == null )
			{
				this.child = task;
			}
			else
			{
				this.child.append( task );
			}
		}
		this.maxId = actors.length;
	};

	/**
	 * 
	 */
	pt.processAction = function( scene )
	{
		for ( var actor = this.child; actor != null; actor = actor.next )
		{
			actor.processAction( scene );
		}
	};
	/**
	 * 
	 */
	pt.processActionByNetwork = function( scene, actors )
	{
		if ( actors == null )
		{
			// process player
			this.child.processAction( scene );
		}
		else
		{
			// make map
			var map = {};
			for ( var i = 0; i < actors.length; i++ )
			{
				var ad = actors[i];
				map[ ad.clientId ] = ad;
			}
			// process
			for ( var actor = this.child; actor != null; actor = actor.next )
			{
				var ad = map[ actor.ad.clientId ];
				// update
				if ( ad )
				{
					actor.processActionByNetwork( scene, ad );
					delete map[ actor.ad.clientId ];
				}
				// remove
				else
				{
					// clear map cell
					scene.stage.map[ Math.floor( actor.ad.y ) ][ Math.floor( actor.ad.x ) ] = 0;
					actor.unlink();
					//if ( actor == this.child )..
				}
			}
			// add
			for ( var key in map )
			{
				var task = new ActorTask( );
				task.init( scene, this.maxId++, map[ key ] );
				this.child.append( task );
			}
		}
	};

	/**
	 * 
	 */
	pt.update = function( scene )
	{
		return this.updateChildren( scene );
	};

	/**
	 * 
	 */
	pt.draw = function( scene )
	{
		this.pickVisible( scene );

		var context		= scene.context;
		var canvas		= context.canvas;
		var resource	= scene.data.resource;
		var stage		= scene.stageTask;
		var viewPoint	= scene.viewPoint;

		for ( var i = 0; i < this.visible.length; i++ )
		{
			var actor = this.visible[i];
			// if behind wall
			if ( stage.rd[ actor.rd.index ].dist <= actor.rd.dist )
			{
				continue;
			}

			var offSx = actor.animator.fx;
			var offSy = 0;
			// angle
			var sdir = viewPoint.ad.dir - actor.ad.dir;
			if ( sdir < 0 ) { sdir += DEG_PI2; }
			sdir /= ( DEG_PI2 / 8 );
			if ( sdir < 1 || sdir >= 7 )
			{
				sdir = 0;
			}
			else
			{
				sdir = Math.floor( ( sdir + 1 ) / 2 );
			}
			offSy = sdir * actor.animator.fh;

			var dh	= canvas.height / actor.rd.dist;
			var dw	= dh;
			var dx	= actor.rd.index * RenderInfo.rayPixels - dw / 2;
			var dy	= canvas.height / 2 - dh / 2;

			var rx	= dx / RenderInfo.rayPixels;
			var rw	= dw / RenderInfo.rayPixels;

			for ( var n = 0; n < rw; n++ )
			{
				var rid = Math.floor( rx + n );
				if ( rid < 0 ) { continue; }
				if ( rid >= RenderInfo.samples ) { break; }
				// draw
				if ( stage.rd[ rid ].dist > actor.rd.dist )
				{
					var image = actor.animator.image;
					var sx = offSx + ( n / rw ) * ( RenderInfo.chipPixels - 1 );
					var sy = offSy;
					var sw = RenderInfo.chipPixels / rw;
					var sh = RenderInfo.chipPixels;
					context.drawImage(
						image, sx, sy, sw, sh, dx + ( n * RenderInfo.rayPixels ), dy, RenderInfo.rayPixels, dh );
				}
			}
		}
	};
	/**
	 * 
	 */
	pt.pickVisible = function( scene )
	{
		var viewPoint	= scene.viewPoint;
		var stage		= scene.stageTask;

		var radian		= viewPoint.getDirRad();
		var startTheta	= radian + ( RenderInfo.fov / 2 );

		// search visible actor
		this.visible.length = 0;
		for ( var actor = this.child.next; actor != null; actor = actor.next )
		{
			var x		= actor.ad.x - viewPoint.ad.x;
			var y		= viewPoint.ad.y - actor.ad.y;
			var thetaV	= Math.atan2( y, x );	// PI ... PI * -1
			if ( thetaV < 0 ) { thetaV += PI2; }

			var diff	= startTheta - thetaV;
			if ( 0 <= diff && diff <= RenderInfo.fov )
			{
				actor.rd.index	= Math.floor( diff / ( RenderInfo.fov / RenderInfo.samples ) );
				actor.rd.dist	= Math.sqrt( x * x + y * y );
				this.visible.push( actor );
			}
		}
		this.visible.sort( function( a, b ) { return b.rd.dist - a.rd.dist;  } );
	};
})( ActorListTask.prototype );

/**
 * ActorStatus
 */
var ActorStatus =
{
	STATIC	: 0,
	ACTIVE	: 1,
	MORTAL	: 2
};

/**
 * ActorTask
 */
var ActorTask = function()
{
	this.animator		= null;
	this.id		= 0;
	// actor data
	this.ad		= null;
	// raycasting data
	this.rd		=
	{
		index	: 0,
		dist	: 0
	};
	// step data
	this.step	=
	{
		goal	: { x : 0, y : 0, dir : -1 },
		count	: 0,
		locked	: false
	}
};
ActorTask.prototype = new Task();

(function( pt )
{
	pt.ROTATE_VAL	= 10;
	pt.STEP_VAL		= 0.32;
	pt.STEP_COUNT	= 4;

	/**
	 * 
	 */
	pt.init = function( scene, id, ad )
	{
		this.id		= id;
		this.ad		= ad;

		this.moveTo( scene.stageTask, this.ad.x, this.ad.y );
		this.command	= ( id == 0 ) ? scene.command : new AICommand();

		var image		= scene.data.resource.image.actors[ this.ad.type ].data;
		var animation	= scene.data.animation;
		this.interfaces	=
		[
			{ update : this.updateActive,	animator : new ImageAnimator( image, animation.actorAction ) },
			{ update : this.updateActive,	animator : new ImageAnimator( image, animation.actorAction ) },
			{ update : this.updateMortal,	animator : new ImageAnimator( image, animation.actorMotal ) }
		];
		this.setStatus( this.ad.status );
	};
	/**
	 * 
	 */
	pt.getDirRad = function()
	{
		return ( this.ad.dir * PI2 / DEG_PI2 );
	};
	/**
	 * 
	 */
	pt.getGoalDir = function( val )
	{
		var dir	= this.ad.dir + val;
		dir		%= DEG_PI2;
		if ( dir < 0 ) { dir += DEG_PI2; }

		return dir;
	};
	/**
	 * 
	 */
	pt.getGoalPos = function( val )
	{
		var radian	= this.getDirRad();
		var pos =
		{
			x : this.ad.x + Math.cos( radian ) * val,
			y : this.ad.y - Math.sin( radian ) * val
		};
		return pos;
	};
	/**
	 * 
	 */
	pt.moveTo = function( stage, x, y )
	{
		// restore map value
		stage.map[ Math.floor( this.ad.y ) ][ Math.floor( this.ad.x ) ] = 0;
		this.ad.x = x;
		this.ad.y = y;
		// update map value
		stage.map[ Math.floor( this.ad.y ) ][ Math.floor( this.ad.x ) ] = Config.actorValBase + this.id;
	};
	/**
	 * 
	 */
	pt.isInnerWall = function( stage, x, y )
	{
		var arr = [ -0.3, 0.3 ];

		for ( var i = 0; i < arr.length; i++ )
		{
			var my = Math.floor( y + arr[i] );
			if ( my < 0 || stage.getMapHeight() <= my ) { return true; }

			for ( var j = 0; j < arr.length; j++ )
			{
				var mx = Math.floor( x + arr[j] );
				if ( mx < 0 || stage.getMapWidth() <= mx ) { return true; }

				var value = stage.map[ my ][ mx ];
				if ( value != 0 && value != Config.actorValBase + this.id )
				{
					return true;
				}
			}
		}
		return false;
	};
	/**
	 * 
	 */
	pt.processAction = function( scene )
	{
		if ( this.status == ActorStatus.ACTIVE && !this.step.locked )
		{
			this.command.update();

			// left right
			if ( this.command.tbl.left )
			{
				this.step.goal.dir	= this.getGoalDir( this.ROTATE_VAL );
			}
			else if ( this.command.tbl.right )
			{
				this.step.goal.dir	= this.getGoalDir( this.ROTATE_VAL * -1 );
			}
			// go back
			else if ( this.command.tbl.go )
			{
				var pos = this.getGoalPos( this.STEP_VAL );
				this.step.goal.x	= pos.x;
				this.step.goal.y	= pos.y;
				this.step.count		= this.STEP_COUNT;
			}
			else if ( this.command.tbl.back )
			{
				var pos = this.getGoalPos( this.STEP_VAL * -1 );
				this.step.goal.x	= pos.x;
				this.step.goal.y	= pos.y;
				this.step.count		= this.STEP_COUNT;
			}
		}
	};
	/**
	 * 
	 */
	pt.processActionByNetwork = function( scene, ad )
	{
		// pos
		if ( Math.floor( this.ad.x ) != Math.floor( ad.x ) || Math.floor( this.ad.y ) != Math.floor( ad.y ) )
		{
			this.step.goal.x	= ad.x;
			this.step.goal.y	= ad.y;
			this.step.count		= this.STEP_COUNT;
			this.step.locked	= true;
		}
		// dir
		if ( this.ad.dir != ad.dir && this != scene.viewPoint )	// player 以外
		{
			this.step.goal.dir	= ad.dir;
		}
		// message
		if ( this.ad.message.time != ad.message.time )
		{
			var task = new TextRunnerTask( scene );
			task.init( scene, ad.message.text );
			task.link( scene.frontViewTask );

			this.ad.message.time	= ad.message.time;
			this.ad.message.text	= ad.message.text;
		}
	};

	/**
	 * 
	 */
	pt.updateActive = function( scene )
	{
		var upd = false;
		var stage	= scene.stageTask;

		// dir
		if ( this.step.goal.dir != -1 )
		{
			this.ad.dir			= this.step.goal.dir;
			this.step.goal.dir	= -1;
			upd = true;
		}
		// step
		if ( this.step.count != 0 )
		{
			this.step.count--;
			if ( this.step.count ) { this.step.locked = false; }
			var newX = ( this.step.count ) ? ( this.ad.x + this.step.goal.x ) / 2 : this.step.goal.x;
			var newY = ( this.step.count ) ? ( this.ad.y + this.step.goal.y ) / 2 : this.step.goal.y;

			if ( !this.isInnerWall( scene.stageTask, newX, this.ad.y ) )
			{
				this.moveTo( stage, newX, this.ad.y );
				upd = true;
			}
			if ( !this.isInnerWall( scene.stageTask, this.ad.x, newY ) )
			{
				this.moveTo( stage, this.ad.x, newY );
				upd = true;
			}
		}
		// animation
		if ( this.animator.isActive() )
		{
			upd = this.animator.proceed();
		}
		else
		{
			var r = Math.floor( Math.random() * 10 );
			if ( r == 1 )
			{
				this.animator.start();
				upd = true;
			}
		}
		return upd;
	};
	/**
	 * 
	 */
	pt.updateMortal = function( scene )
	{
		var upd = false;
		// animation
		if ( this.animator.isActive() )
		{
			upd = this.animator.proceed();
		}
		return upd;
	};
})( ActorTask.prototype );

/**
 * EventTask
 */
var EventTask = function( scene )
{
	this.command	= scene.command;	// input
	this.prevDist	= [];
};
EventTask.prototype = new Task();

(function( pt )
{
	pt.EVENT_DIST	= 1.1;
	pt.FIRE_DIST	= 2;

	/**
	 * 
	 */
	pt.update = function( scene )
	{
		var upd			= false;
		var actorList	= scene.actorListTask;
		var msgManager	= scene.msgManager;
		var resource	= scene.data.resource;
		var sound		= scene.soundTask;

		// check distance
		for ( var i = 0; i < actorList.visible.length; i++ )
		{
			var actor = actorList.visible[i];
			// check event
			if ( actor.rd.dist < this.EVENT_DIST )
			{
				var dist = this.prevDist[ actor.id ];
				if ( dist != undefined && dist > this.EVENT_DIST )
				{
					// text box
					if ( actor.ad.event.type == EventType.TEXT )
					{
						sound.prepare( resource.sound.notice.data );

						// add task
						var task = new TextBoxTask( scene );
						task.link( this );
						task.eventActor	= actor;
						// setFocus
						scene.setFocus( task );

						upd = true;
					}
					// battle
					else if ( actor.ad.event.type == EventType.BATTLE )
					{
						var message = new Message( MessageType.START_BATTLE );
						message.data.enemy = actor;
						msgManager.postBreakMessage( message );

						upd = true;
					}
					// warp
					else if ( actor.ad.event.type == EventType.WARP )
					{
						sound.prepare( resource.sound.warp.data );

						var message = new Message( MessageType.WARP_SCENE );
						message.data.baseUrl = actor.ad.event.baseUrl;
						msgManager.postBreakMessage( message );

						upd = true;
					}
					break;
				}
			}
			// check fire
			if ( this.command.tbl.fire )
			{
				if ( actor.rd.dist < this.FIRE_DIST && actor.status == ActorStatus.ACTIVE )
				{
					actor.setStatus( ActorStatus.MORTAL );
					actor.animator.start();
					upd = true;
					break;
				}
			}
		}
		// copy dist
		this.prevDist.length = 0;
		for ( var i = 0; i < actorList.visible.length; i++ )
		{
			var actor = actorList.visible[i];
			this.prevDist[ actor.id ] = actor.rd.dist;
		}

		// 画面遷移
		if ( this.command.tbl.editor )
		{
			var message = new Message( MessageType.START_EDITOR );
			msgManager.postBreakMessage( message );
		}
		else if ( this.command.tbl.escape )
		{
			var message	= new Message( MessageType.RET_PREV_SCENE );
			msgManager.postBreakMessage( message );
		}
		return upd;
	};
})( EventTask.prototype );

/**
 * TextBoxTask
 */
var TextBoxTask = function( scene )
{
	this.command	= scene.command;	// input
	this.eventActor	= null;

	this.lag		= this.LAG_VAL;
};
TextBoxTask.prototype = new Task();

(function( pt )
{
	pt.LAG_VAL	= 10;

	/**
	 * 
	 */
	pt.update = function( scene )
	{
		var upd = false;

		if ( scene.getFocus() == this )
		{
			if ( this.lag > 0 )
			{
				this.lag--;
			}
			else if ( this.command.tbl.go || this.command.tbl.back || this.command.tbl.left || this.command.tbl.right )
			{
				scene.setFocus( null );
				this.unlink();
				upd = true;
			}
		}
		return upd;
	};

	/**
	 * 
	 */
	pt.draw = function( scene )
	{
		var context		= scene.context;
		var canvas		= context.canvas;

		var tx = 10;
		var ty = 10;
		var th = 50;

		context.save();

		// rect
		context.fillStyle	= Res.Color.TEXTBOX;
		context.fillRect( tx, ty, canvas.width - tx * 2, th );
		// border
		context.strokeStyle	= Res.Color.BORDER_TEXTBOX;
		context.beginPath();
		context.strokeRect( tx, ty, canvas.width - tx * 2, th );
		// text
		context.font		= Res.Font.LARGE;
		context.fillStyle	= Res.Color.FONT_TEXTBOX;
		context.fillText( this.eventActor.ad.event.text, tx + 10, ty + 25 );

		context.restore();
	};
})( TextBoxTask.prototype );

/**
 * FireTask
 */
var FireTask = function( scene, cx, cy )
{
	this.animator	= new ImageAnimator( scene.data.resource.image.fire.data, scene.data.animation.fire );
	this.animator.start();
	this.trans	= { x : cx, y : cy };	// center position
};
FireTask.prototype = new Task();

(function( pt )
{
	/**
	 * 
	 */
	pt.update = function( scene )
	{
		var upd = false;

		if ( this.animator.isActive() )
		{
			upd = this.animator.proceed();
		}
		else
		{
			this.unlink();
			upd = true;
		}
		return upd;
	};

	/**
	 * 
	 */
	pt.draw = function( scene )
	{
		var context		= scene.context;

		context.save();

		context.drawImage(
			this.animator.image,
			this.animator.fx,							// sx
			this.animator.fy,							// sy
			this.animator.fw,							// sw
			this.animator.fh,							// sh
			this.trans.x - ( this.animator.fw / 2 ),	// dx
			this.trans.y - ( this.animator.fh / 2 ),	// dy
			this.animator.fw,							// dw
			this.animator.fh							// dh
		);

		context.restore();
	};
})( FireTask.prototype );

/**
 * FrontViewTask
 */
var FrontViewTask = function( scene )
{
	this.command	= scene.command;	// input
	this.gx			= 0;
	this.gy			= 0;
	this.lag		= 0;
};
FrontViewTask.prototype = new Task();

(function( pt )
{
	pt.LAG_VAL	= 5;

	/**
	 * 
	 */
	pt.update = function( scene )
	{
		var upd			= false;
		var resource	= scene.data.resource;
		var sound		= scene.soundTask;

		// lag
		if ( this.lag )
		{
			this.lag--;
			upd = true;
		}
		else
		{
			if ( this.command.tbl.fire )
			{
				sound.prepare( resource.sound.shoot.data );

				var task	= new FireTask( scene, this.gx, this.gy );
				task.link( this.prev );
				this.lagF	= this.LAG_VAL;
				upd = true;
			}
		}
		return upd;
	};

	/**
	 * 
	 */
	pt.draw = function( scene )
	{
		var context		= scene.context;
		var canvas		= context.canvas;
		var resource	= scene.data.resource;
		var stage		= scene.stageTask;
		var viewPoint	= scene.viewPoint;
		var image;

		// gun
		this.gx = canvas.width  - resource.image.gun.data.width;
		this.gy = canvas.height - resource.image.gun.data.height;
		this.gx += 20 + Math.cos( scene.ticks / 6 ) * viewPoint.step.count * 4;
		this.gy += 20 + Math.cos( scene.ticks / 5 ) * viewPoint.step.count * 4;

		image = resource.image.gun.data;
		context.drawImage(
			image,
			0,				// sx
			0,				// sy
			image.width,	// sw
			image.height,	// sh
			this.gx,		// dx
			this.gy,		// dy
			image.width,	// dw
			image.height	// dh
		);
	};
})( FrontViewTask.prototype );

/**
 * NetworkTask
 */
var NetworkTask = function( scene )
{
	this.command		= scene.command;	// input
	this.lastTime		= (new Date()).getTime();
	this.loginTime		= 0;
	this.clientId		= null;
	this.tmpActors		= null;
	this.messageText	= null;
};
NetworkTask.prototype = new Task();

(function( pt )
{
	/**
	 * 
	 */
	pt.sortActors = function( actors )
	{
		var newActors = [];
		var flg = false;
		for ( var i = 0; i < actors.length; i++ )
		{
			var ad = actors[i];
			if ( ad.clientId == this.clientId )
			{
				newActors.unshift( ad );
				flg = true;
			}
			else
			{
				newActors.push( ad );
			}
		}
		return ( flg ) ? newActors : null;
	}


	/**
	 * 
	 */
	pt.joinServer = function( scene )
	{
		// update lastTime
		this.lastTime	= (new Date()).getTime();
		this.loginTime	= this.lastTime;

		this.clientId	= this.lastTime + "_" + Math.floor( Math.random() * 10000 );

		var self	= this;
		var request	= scene.app.createXMLHttpRequest();
		request.open( "POST", ProtocolInfo.url, true );
		request.setRequestHeader( "Content-Type", "application/x-www-form-urlencoded" );
		request.onload = function( event )
		{
			try
			{
				self.lastTime	= (new Date()).getTime();
				self.loginTime	= self.lastTime;

				var data = eval( "(" + request.responseText + ")" );
				if ( data.map && data.actors )
				{
					// sort
					data.actors	= self.sortActors( data.actors );
					if ( data.actors == null )
					{
						Debug.alert( Res.String.TM_JOIN_NG );
						self.backToMenu( scene );
					}
					// set data
					scene.data.map		= data.map;
					scene.data.actors	= data.actors;
					scene.setData( data );
					// set status
					scene.setStatus( SceneStatus.RUNNING );
					scene.show();
				}
				// NG
				else
				{
					Debug.alert( Res.String.TM_JOIN_NG + "(" + data.errorMessage + ")" );
					self.backToMenu( scene );
				}
			}
			catch ( e )
			{
				Debug.alertError( e );
				self.backToMenu( scene );
			}
		};
		request.onerror = function( event )
		{
			Debug.alert( Res.String.TM_CONN_ERR + ", " + event.target.status );
			self.backToMenu( scene );
		};
		// send
		var query =
			"method="		+ window.encodeURIComponent( ProtocolInfo.method.join ) +
			"&clientId="	+ window.encodeURIComponent( this.clientId );
		request.send( query );
	};
	/**
	 * 
	 */
	pt.updateServer = function( scene )
	{
		// update lastTime
		this.lastTime	= (new Date()).getTime();

		var viewPoint	= scene.viewPoint;

		var self	= this;
		var request	= scene.app.createXMLHttpRequest();
		request.open( "POST", ProtocolInfo.url, true );
		request.setRequestHeader( "Content-Type", "application/x-www-form-urlencoded" );
		request.onload = function( event )
		{
			try
			{
				self.lastTime	= (new Date()).getTime();

				var data = eval( "(" + request.responseText + ")" );
				if ( data.actors )
				{
					// sort
					data.actors	= self.sortActors( data.actors );
					if ( data.actors == null )
					{
						Debug.alert( Res.String.TM_UPDATE_NG );
						self.backToMenu( scene );
					}
					self.tmpActors = data.actors;
				}
				// NG
				else
				{
					Debug.alert( Res.String.TM_UPDATE_NG + "(" + data.errorMessage + ")" );
					self.backToMenu( scene );
				}
			}
			catch ( e )
			{
				Debug.alertError( e );
				self.backToMenu( scene );
			}
		};
		request.onerror = function( event )
		{
			Debug.alert( Res.String.TM_CONN_ERR + ", " + event.target.status );
			self.backToMenu( scene );
		};
		// send
		var query =
			"method="		+ window.encodeURIComponent( ProtocolInfo.method.update ) +
			"&clientId="	+ window.encodeURIComponent( this.clientId ) +
			"&x="			+ window.encodeURIComponent( viewPoint.ad.x ) +
			"&y="			+ window.encodeURIComponent( viewPoint.ad.y ) +
			"&dir="			+ window.encodeURIComponent( viewPoint.ad.dir );
		if ( this.messageText != null )
		{
			query += "&messageText=" + window.encodeURIComponent( this.messageText );
			this.messageText	= null;
		}
		request.send( query );
	}
	/**
	 * 
	 */
	pt.backToMenu = function( scene )
	{
		var message	= new Message( MessageType.RET_PREV_SCENE );
		scene.msgManager.postBreakMessage( message );
	}

	/**
	 * 
	 */
	pt.update = function( scene )
	{
		var upd			= false;
		var actorList	= scene.actorListTask

		if ( scene.online )
		{
			actorList.processActionByNetwork( scene, this.tmpActors );
			if ( this.tmpActors != null )
			{
				this.tmpActors	= null;
			}

			var nowTime = (new Date()).getTime();
			// logout
			if ( ( nowTime - this.loginTime ) >  ProtocolInfo.playTimeMax )
			{
				this.backToMenu( scene );
				return true;
			}
			// send update
			if ( ( nowTime - this.lastTime ) >  ProtocolInfo.updateInterval )
			{
				this.updateServer( scene );
			}
		}
		else
		{
			actorList.processAction( scene );
		}
		return upd;
	};
})( NetworkTask.prototype );

/**
 * TextRunnerTask
 */
var TextRunnerTask = function( scene )
{
	this.command	= scene.command;	// input
	this.text		= null;

	this.trans		= { x : 0, y : 0 };	// left, top
};
TextRunnerTask.prototype = new Task();

(function( pt )
{
	pt.STEP_VAL	= 20;

	/**
	 * 
	 */
	pt.init = function( scene, text )
	{
		var context		= scene.context;
		var canvas		= context.canvas;

		this.trans.x	= canvas.width - 10;
		this.trans.y	= Math.floor( Math.random() * ( canvas.height - 80 ) );

		this.text		= text;
	};
	/**
	 * 
	 */
	pt.update = function( scene )
	{
		this.trans.x -= this.STEP_VAL;
		if ( this.trans.x < -50 )	//...要修正
		{
			this.unlink();
		}
		return true;
	};
	/**
	 * 
	 */
	pt.draw = function( scene )
	{
		var context		= scene.context;

		context.save();

		// text
		context.font		= Res.Font.EXLARGE;
		context.fillStyle	= Res.Color.FONT_TEXTBOX;
		context.fillText( this.text, this.trans.x, this.trans.y + 30 );

		context.restore();
	};
})( TextRunnerTask.prototype );

/**
 * DebugTask
 */
var DebugTask = function( scene )
{
	this.command	= scene.command;	// input
	this.info		= null;
	this.lag		= 0;
};
DebugTask.prototype = new Task();

(function( pt )
{
	pt.LAG_VAL	= 10;

	/**
	 * 
	 */
	pt.update = function( scene )
	{
		var upd = false;
		// lag
		if ( this.lag )
		{
			this.lag--;
		}
		else
		{
			if ( this.command.tbl.debug )
			{
				this.info	= "viewPoint.ad.x=" + scene.viewPoint.ad.x + ", viewPoint.ad.y=" + scene.viewPoint.ad.y;
				this.lag	= this.LAG_VAL;
				upd = true;
			}
		}
		return upd;
	};

	/**
	 * 
	 */
	pt.draw = function( scene )
	{
		// output debug info
		if ( this.info )
		{
			Debug.print( this.info );
			this.info = null;
		}
	};
})( DebugTask.prototype );


/**
 * AICommand
 */
var AICommand = function()
{
	this.tbl.left	= 0;
	this.tbl.right	= 0;
	this.tbl.go		= 0;
	this.tbl.back	= 0;
};
AICommand.prototype = new Command();

(function( pt )
{
	/**
	 * 
	 */
	pt.update = function()
	{
		this.clear();

		var r = Math.floor( Math.random() * 18 );
		if      ( 0 <= r && r <= 3 )	{ this.tbl.left		= 1; }
		else if ( 4 <= r && r <= 8 )	{ this.tbl.right	= 1; }
		else if ( r == 9 )				{ this.tbl.go		= 1; }
		else if ( r == 10 )				{ this.tbl.back		= 1; }
	};
})( AICommand.prototype );

/**
 * InputCommand
 */
var InputCommand = function()
{
	this.tbl.left	= 0;
	this.tbl.right	= 0;
	this.tbl.go		= 0;
	this.tbl.back	= 0;
	this.tbl.fire	= 0;
	this.tbl.seek	= 0;
	this.tbl.editor	= 0;
	this.tbl.escape	= 0;
	this.tbl.debug	= 0;
};
InputCommand.prototype = new Command();

(function( pt )
{
	/**
	 * 
	 */
	pt.update = function() {};
	/**
	 * 
	 */
	pt.handleSysEvent = function( event )
	{
		var type = event.type.toLowerCase();
		if ( type.substring( 0, 3 ) == "key" )
		{
			var value = 0;
			if      ( type == "keydown" ) { value = 1; }
			else if ( type == "keyup"   ) { value = 0; }

			switch ( event.keyCode )
			{
				case 37: this.tbl.left	= value; event.prevent( event.orgEvent ); break;	// ←
				case 39: this.tbl.right	= value; event.prevent( event.orgEvent ); break;	// →
				case 38: this.tbl.go	= value; event.prevent( event.orgEvent ); break;	// ↑
				case 40: this.tbl.back	= value; event.prevent( event.orgEvent ); break;	// ↓
				/*
				case 65: this.tbl.left	= value; break;	// A
				case 68: this.tbl.right	= value; break;	// D
				case 87: this.tbl.go	= value; break;	// W
				case 83: this.tbl.back	= value; break;	// S
				*/
				case 90: this.tbl.fire	= value; break;	// Z
				case 32: this.tbl.seek	= value; break;	// space;
				case 69: this.tbl.editor= value; break;	// E
				case 27: this.tbl.escape= value; break;	// ESC
				case 81: this.tbl.debug	= value; break;	// Q
			}
		}
	};
})( InputCommand.prototype );

/**
 * StageScene
 */
var StageScene = function( app, name, online )
{
	this.app			= app;
	this.name			= name;
	this.online			= online;
	this.data			= null;

	this.command		= new InputCommand();
	this.viewPoint		= null;

	// create task
	this.networkTask	= new NetworkTask( this );
	this.stageTask		= new StageTask( this );
	this.actorListTask	= new ActorListTask( this );
	this.eventTask		= new EventTask( this );
	this.frontViewTask	= new FrontViewTask( this );
	this.soundTask		= new SoundTask( this );
	this.debugTask		= new DebugTask( this );
	// create list
	this.networkTask.append( this.stageTask );
	this.networkTask.append( this.actorListTask );
	this.networkTask.append( this.eventTask );
	this.networkTask.append( this.frontViewTask );
	this.networkTask.append( this.soundTask );
	this.networkTask.append( this.debugTask );
	// head of task list
	this.child		= this.networkTask;
	this.setStatus( SceneStatus.READY );

	// message handler
	this.msgManager		= new MessageManager();
	this.msgManager.addHandler( new MessageHandler( MessageType.RET_PREV_SCENE,	this, this.handleRetPrevScene ) );
	this.msgManager.addHandler( new MessageHandler( MessageType.START_BATTLE,	this, this.handleStartBattle ) );
	this.msgManager.addHandler( new MessageHandler( MessageType.WARP_SCENE,		this, this.handleWarpScene ) );
	this.msgManager.addHandler( new MessageHandler( MessageType.START_EDITOR,	this, this.handleStartEditor ) );
};
StageScene.prototype = new Scene();

(function( pt )
{
	/**
	 * 
	 */
	pt.handleRetPrevScene = function( scene, message )
	{
		this.app.sceneManager.pop();
		var prevScene = this.app.sceneManager.current;
		prevScene.show();
	};
	/**
	 * 
	 */
	pt.handleStartBattle = function( scene, message )
	{
		var battleScene	= this.app.sceneManager.push( Res.String.BATTLE_NAME );
		battleScene.reset( this.data, this.viewPoint, message.data.enemy );
		battleScene.show();
	};
	/**
	 * 
	 */
	pt.handleWarpScene = function( scene, message )
	{
		// set status
		this.setStatus( SceneStatus.READY );
		// pop this
		this.app.sceneManager.pop();

		var baseUrl	= message.data.baseUrl;
		var stageScene	= this.app.sceneManager.push( baseUrl );
		if ( stageScene )
		{
			// set status
			stageScene.setStatus( SceneStatus.RUNNING );
			stageScene.show();	// clear input
		}
		else
		{
			// create scene
			stageScene	= new StageScene( this.app, baseUrl );

			this.app.sceneManager.add( stageScene );
			this.app.sceneManager.push( stageScene.name );

			stageScene.init();
			stageScene.show();

			// load data
			var base = ( baseUrl.charAt( baseUrl.length - 1 ) == '/' ) ? baseUrl : baseUrl + '/';
			this.app.loadScript( base + Config.dataJSName );
		}
	};
	/**
	 * 
	 */
	pt.handleStartEditor = function( scene, message )
	{
		var editorScene	= this.app.sceneManager.push( Res.String.EDITOR_NAME );
		editorScene.setData( this.data );
		// set status
		editorScene.setStatus( SceneStatus.RUNNING );
		editorScene.show();
	};

	/**
	 * 
	 */
	pt.init = function()
	{
	};
	/**
	 * 
	 */
	pt.show = function()
	{
		var div	= this.app.getDomElement( Res.String.ID_MESSAGE );
		if ( div )
		{
			if ( this.online )
			{
				div.innerHTML = Res.String.INPUT_HTML;
				// add handler
				var self	= this;
				var button	= this.app.getDomElement( Res.String.ID_SEND );
				if ( button )
				{
					button.onclick = function()
					{
						try
						{
							var text	= self.app.getDomElement( Res.String.ID_TEXT );
							self.networkTask.messageText = text.value;
							text.value = "";
						}
						catch ( e )
						{
							self.app.kill();
							Debug.alertError( e );
						}
					};
				}
			}
			else
			{
				div.innerHTML = "";
			}
		}

		this.setUsage( ( this.online ) ? Res.String.HTML_USAGE_N : Res.String.HTML_USAGE );

		this.command.clear();
		this.holdContext();
		this.draw( this );
	};

	/**
	 * 
	 */
	pt.setData = function( data )
	{
		if ( !this.data )
		{
			this.data	= data;
		}

		if ( data.map )
		{
			// load stage
			this.stageTask.setData( this, data.map );
		}
		if ( data.actors )
		{
			// load actorList
			this.actorListTask.setData( this, data.actors );
			// init viewPoint
			this.viewPoint	= this.actorListTask.child;
		}
	};
	/**
	 * 
	 */
	pt.getData = function()
	{
		return this.data;
	};
	/**
	 * 
	 */
	pt.loadNetworkData = function()
	{
		this.networkTask.joinServer( this );
	};
	/**
	 * 
	 */
	pt.loadData = function( data )
	{
		try
		{
			var self = this;

			// load resource
			this.app.loadResource( this.name, data.resource );
			// set data
			this.setData( data );

			if ( this.online )
			{
				//self.loadNetworkData();
				window.setTimeout( function() { self.loadNetworkData(); }, Config.loadInterval );
			}
			else
			{
				// set status
				window.setTimeout( function() { self.setStatus( SceneStatus.RUNNING ); self.show(); }, Config.loadInterval + 300 );
			}
		}
		catch ( e )
		{
			this.app.kill();
			Debug.alertError( e );
		}
	};
})( StageScene.prototype );


// Expose
if ( !window.h5glib ) { window.h5glib = {}; }
window.h5glib.StageScene	= StageScene;

// Anonymous function end
//
})( window );
